/*-
 * #START_LICENSE#
 * smp-server-library
 * %%
 * Copyright (C) 2017 - 2024 European Commission | eDelivery | DomiSMP
 * %%
 * Licensed under the EUPL, Version 1.2 or – as soon they will be approved by the European Commission - subsequent
 * versions of the EUPL (the "Licence");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * [PROJECT_HOME]\license\eupl-1.2\license.txt or https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the Licence is
 * distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and limitations under the Licence.
 * #END_LICENSE#
 */
package eu.europa.ec.edelivery.smp.services.ui;

import eu.europa.ec.edelivery.smp.data.ui.CertificateRO;
import eu.europa.ec.edelivery.smp.exceptions.CertificateNotTrustedException;
import eu.europa.ec.edelivery.smp.exceptions.ErrorCode;
import eu.europa.ec.edelivery.smp.exceptions.SMPRuntimeException;
import eu.europa.ec.edelivery.smp.services.AbstractServiceIntegrationTest;
import eu.europa.ec.edelivery.smp.services.CRLVerifierService;
import eu.europa.ec.edelivery.smp.services.ConfigurationService;
import eu.europa.ec.edelivery.smp.testutil.X509CertificateTestUtils;
import org.apache.commons.io.FileUtils;
import org.apache.commons.io.IOUtils;
import org.hamcrest.CoreMatchers;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentMatchers;
import org.mockito.Mockito;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.test.util.ReflectionTestUtils;

import javax.security.auth.x500.X500Principal;
import java.io.File;
import java.io.IOException;
import java.math.BigInteger;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.security.cert.*;
import java.time.OffsetDateTime;
import java.util.*;

import static org.hamcrest.CoreMatchers.containsString;
import static org.hamcrest.MatcherAssert.assertThat;
import static org.junit.jupiter.api.Assertions.*;

public class UITruststoreServiceIntegrationTest extends AbstractServiceIntegrationTest {

    public static final String CERTIFICATE_POLICY_ANY = "2.5.29.32.0";
    public static final String CERTIFICATE_POLICY_QCP_NATURAL = "0.4.0.194112.1.0";
    public static final String CERTIFICATE_POLICY_QCP_LEGAL = "0.4.0.194112.1.1";
    public static final String CERTIFICATE_POLICY_QCP_NATURAL_QSCD = "0.4.0.194112.1.2";
    public static final String CERTIFICATE_POLICY_QCP_LEGAL_QSCD = "0.4.0.194112.1.3";

    public static final String S_SUBJECT_PEPPOL = "CN=POP000004,OU=PEPPOL TEST AP,O=European Commission,C=BE";
    public static final String S_SUBJECT_PEPPOL_EXPANDED = "serialNumber=12345,emailAddress=test@mail.com,CN=POP000004,OU=PEPPOL TEST AP,O=European Commission,street=My Street,C=BE";
    public static final String S_SUBJECT_PEPPOL_NOT_TRUSTED = "CN=POP000005,OU=PEPPOL TEST AP,O=European Commission,C=BE";

    public static final String S_SUBJECT_TEST = "CN=SMP test,O=DIGIT,C=BE";


    Path resourceDirectory = Paths.get("src", "test", "resources", "truststore");
    Path targetDirectory = Paths.get("target", "truststore");

    @Autowired
    protected UITruststoreService testInstance;

    @Autowired
    protected ConfigurationService configurationService;

    @Autowired
    protected CRLVerifierService crlVerifierService;

    @BeforeEach
    public void setup() throws IOException {
        configurationService = Mockito.spy(configurationService);
        crlVerifierService = Mockito.spy(crlVerifierService);
        testInstance = Mockito.spy(testInstance);


        ReflectionTestUtils.setField(testInstance, "crlVerifierService", crlVerifierService);
        ReflectionTestUtils.setField(testInstance, "configurationService", configurationService);

        File truststoreFile = new File(targetDirectory.toFile(), "smp-truststore.jks");
        Mockito.doReturn("test123").when(configurationService).getTruststoreCredentialToken();
        Mockito.doReturn(truststoreFile).when(configurationService).getTruststoreFile();
        Mockito.doReturn("JKS").when(configurationService).getTruststoreType();
        Mockito.doReturn(targetDirectory.toFile()).when(configurationService).getSecurityFolder();
        Mockito.doReturn(true).when(configurationService).forceCRLValidation();
        resetKeystore();

        testInstance.refreshData();
    }

    public void resetKeystore() throws IOException {
        FileUtils.deleteDirectory(targetDirectory.toFile());
        FileUtils.copyDirectory(resourceDirectory.toFile(), targetDirectory.toFile());
    }

    @Test
    void testGetKeystoreEntriesList() {
        List<String> lst = testInstance.getNormalizedTrustedList();
        assertEquals(2, lst.size());
        assertEquals(S_SUBJECT_PEPPOL, lst.get(0));
        assertEquals(S_SUBJECT_TEST, lst.get(1));
    }

    @Test
    void testSubjectValid() {
        // given when
        // then
        assertTrue(testInstance.isSubjectOnTrustedList(S_SUBJECT_PEPPOL));
    }

    @Test
    void testSubjectValidExpanded() {
        // given when
        // then
        assertTrue(testInstance.isSubjectOnTrustedList(S_SUBJECT_PEPPOL_EXPANDED));
    }

    @Test
    void testSubjectNotTrusted() {
        // given when
        // then
        assertFalse(testInstance.isSubjectOnTrustedList(S_SUBJECT_PEPPOL_NOT_TRUSTED));
    }

    @Test
    void testAddCertificate() throws Exception {
        // given
        String certSubject = "CN=SMP Test,OU=eDelivery,O=DIGITAL,C=BE";
        String alias = UUID.randomUUID().toString();
        X509Certificate certificate = X509CertificateTestUtils.createX509CertificateForTest(certSubject);
        int iSize = testInstance.getNormalizedTrustedList().size();
        assertFalse(testInstance.isSubjectOnTrustedList(certSubject));
        // when
        testInstance.addCertificate(alias, certificate);

        // then
        assertEquals(iSize + 1, testInstance.getNormalizedTrustedList().size());
        assertTrue(testInstance.isSubjectOnTrustedList(certSubject));
    }

    @Test
    void testAddCertificateRDN() throws Exception {
        // given
        String certSubject = "GIVENNAME=John+SERIALNUMBER=1+CN=SMP Test,OU=eDelivery,O=DIGITAL,C=BE";
        String alias = UUID.randomUUID().toString();
        X509Certificate certificate = X509CertificateTestUtils.createX509CertificateForTest(certSubject);
        String val = certificate.getSubjectX500Principal().getName(X500Principal.RFC2253);
        int iSize = testInstance.getNormalizedTrustedList().size();
        assertFalse(testInstance.isSubjectOnTrustedList(certSubject));
        // when
        testInstance.addCertificate(alias, certificate);

        // then
        assertEquals(iSize + 1, testInstance.getNormalizedTrustedList().size());
        assertTrue(testInstance.isSubjectOnTrustedList(certSubject));
    }

    @Test
    void testDeleteCertificate() throws Exception {
        // given
        List<CertificateRO> list = testInstance.getCertificateROEntriesList();
        int iSize = list.size();
        assertTrue(list.size() > 0);
        CertificateRO certificateRO = list.get(0);
        assertTrue(testInstance.isSubjectOnTrustedList(certificateRO.getSubject()));
        // when
        testInstance.deleteCertificate(certificateRO.getAlias());

        // then
        assertEquals(iSize - 1, testInstance.getNormalizedTrustedList().size());
        assertFalse(testInstance.isSubjectOnTrustedList(certificateRO.getSubject()));
    }

    @Test
    void testDeleteCertificateNotExists() throws Exception {
        // given
        // when
        X509Certificate cert = testInstance.deleteCertificate("alias-not-exists");
        // then
        assertNull(cert);
    }

    @Test
    void testIsTruststoreChanged() throws Exception {
        // given
        String certSubject = "CN=SMP Test,OU=eDelivery,O=DIGITAL,C=BE";
        String alias = UUID.randomUUID().toString();
        X509Certificate certificate = X509CertificateTestUtils.createX509CertificateForTest(certSubject);
        testInstance.addCertificate(alias, certificate);
        assertTrue(testInstance.isSubjectOnTrustedList(certSubject));
        // when rollback truststore
        resetKeystore();
        // then it should detect file change and refresh certificates
        assertFalse(testInstance.isSubjectOnTrustedList(certSubject));
    }


    @Test
    void testGetCertificateDataPEMAndFullValidationExpired() throws IOException {
        // given

        byte[] buff = IOUtils.toByteArray(UIUserServiceIntegrationTest.class.getResourceAsStream("/truststore/SMPtest.crt"));

        // when
        CertificateRO cer = testInstance.getCertificateData(buff, true, true);

        //then
        assertEquals("CN=SMP test,O=DIGIT,C=BE:0000000000000003", cer.getCertificateId());
        assertEquals("CN=Intermediate CA,O=DIGIT,C=BE", cer.getIssuer());
        assertEquals("1.2.840.113549.1.9.1=#160c736d7040746573742e636f6d,CN=SMP test,O=DIGIT,C=BE", cer.getSubject());
        assertEquals("3", cer.getSerialNumber());
        assertNotNull(cer.getValidFrom());
        assertNotNull(cer.getValidTo());
        assertTrue(cer.getValidFrom().isBefore(cer.getValidTo()));
        assertEquals("Certificate is expired!", cer.getInvalidReason());
    }

    @Test
    void validateCertificateWithTruststoreNullCertificate() {

        CertificateException result = assertThrows(CertificateException.class,
                () -> testInstance.validateCertificateWithTruststore(null));

        assertThat(result.getMessage(), containsString("The X509Certificate is null "));
    }

    @Test
    void validateCertificateWithTruststoreNullTruststore() throws Exception {
        String certSubject = "CN=SMP Test,OU=eDelivery,O=DIGITAL,C=BE";
        X509Certificate certificate = X509CertificateTestUtils.createX509CertificateForTest(certSubject);
        Mockito.doReturn(null).when(testInstance).getTrustStore();
        // no error thrown
        testInstance.validateCertificateWithTruststore(certificate);
        // invoked once
        Mockito.verify(testInstance, Mockito.times(1)).getTrustStore();
    }

    @Test
    void testGetCertificateDataError() throws IOException, CertificateException {
        // given

        byte[] buff = IOUtils.toByteArray(UIUserServiceIntegrationTest.class.getResourceAsStream("/certificates/cert-not-parsable.pem"));

        // when
        CertificateRO cer = testInstance.getCertificateData(buff);

        //then
        assertTrue(cer.isInvalid());
        assertEquals("Can not read the certificate!", cer.getInvalidReason());
    }

    @Test
    void testGetCertificateDataPEMWithHeader() throws IOException, CertificateException {
        // given
        byte[] buff = IOUtils.toByteArray(UIUserServiceIntegrationTest.class.getResourceAsStream("/truststore/pem-with-header.crt"));

        // when
        CertificateRO cer = testInstance.getCertificateData(buff);

        //then
        assertEquals("CN=alice,O=www.freelan.org,C=FR:0000000000000001", cer.getCertificateId());
        assertEquals("1.2.840.113549.1.9.1=#1613636f6e7461637440667265656c616e2e6f7267,CN=Freelan Sample Certificate Authority,OU=freelan,O=www.freelan.org,L=Strasbourg,ST=Alsace,C=FR", cer.getIssuer());
        assertEquals("1.2.840.113549.1.9.1=#1613636f6e7461637440667265656c616e2e6f7267,CN=alice,OU=freelan,O=www.freelan.org,ST=Alsace,C=FR", cer.getSubject());
        assertEquals("1", cer.getSerialNumber());
        assertNotNull(cer.getValidFrom());
        assertNotNull(cer.getValidTo());
        assertTrue(cer.getValidFrom().isBefore(cer.getValidTo()));
    }

    @Test
    void testGetCertificateDataSMime() throws IOException, CertificateException {
        // given
        byte[] buff = IOUtils.toByteArray(UIUserServiceIntegrationTest.class.getResourceAsStream("/certificates/cert-smime.pem"));

        // when
        CertificateRO cer = testInstance.getCertificateData(buff);

        //then
        assertEquals("CN=edelivery_sml,O=European Commission,C=BE:3cfe6b37e4702512c01e71f9b9175464", cer.getCertificateId());
        assertEquals("CN=PEPPOL SERVICE METADATA PUBLISHER TEST CA - G2,OU=FOR TEST ONLY,O=OpenPEPPOL AISBL,C=BE", cer.getIssuer());
        assertEquals("C=BE,O=European Commission,OU=PEPPOL TEST SMP,CN=edelivery_sml", cer.getSubject());
        assertEquals("3cfe6b37e4702512c01e71f9b9175464", cer.getSerialNumber());
        assertNotNull(cer.getValidFrom());
        assertNotNull(cer.getValidTo());
        assertTrue(cer.getValidFrom().isBefore(cer.getValidTo()));
    }

    @Test
    void testGetCertificateDataDER() throws IOException, CertificateException {
        // given
        byte[] buff = IOUtils.toByteArray(UIUserServiceIntegrationTest.class.getResourceAsStream("/truststore/NewPeppolAP.crt"));

        // when
        CertificateRO cer = testInstance.getCertificateData(buff);

        //then
        assertEquals("CN=POP000004,O=European Commission,C=BE:474980c51478cf62761667461aef5e8e", cer.getCertificateId());
        assertEquals("CN=PEPPOL ACCESS POINT TEST CA - G2,OU=FOR TEST ONLY,O=OpenPEPPOL AISBL,C=BE", cer.getIssuer());
        assertEquals("C=BE,O=European Commission,OU=PEPPOL TEST AP,CN=POP000004", cer.getSubject());
        assertEquals("474980c51478cf62761667461aef5e8e", cer.getSerialNumber());
        assertNotNull(cer.getValidFrom());
        assertNotNull(cer.getValidTo());
        assertTrue(cer.getValidFrom().isBefore(cer.getValidTo()));
    }

    @Test
    void testCheckFullCertificateValidityNotYetValid() throws Exception {
        // given
        String certSubject = "CN=SMP Test,OU=eDelivery,O=DIGITAL,C=BE";
        X509Certificate certificate = X509CertificateTestUtils.createX509CertificateForTest("10af", certSubject, certSubject,
                OffsetDateTime.now().plusDays(1),
                OffsetDateTime.now().plusDays(1),
                Collections.emptyList());
        // when
        assertThrows(CertificateNotYetValidException.class, () ->
                testInstance.checkFullCertificateValidity(certificate));

    }

    @Test
    void testCheckFullCertificateValidityExpired() throws Exception {
        // given
        String certSubject = "CN=SMP Test,OU=eDelivery,O=DIGITAL,C=BE";

        X509Certificate certificate = X509CertificateTestUtils.createX509CertificateForTest(
                "10af", certSubject, certSubject,
                OffsetDateTime.now().minusDays(2),
                OffsetDateTime.now().minusDays(1),
                Collections.emptyList());

        // when
        assertThrows(CertificateExpiredException.class, () ->
                testInstance.checkFullCertificateValidity(certificate));

    }

    @Test
    void testCheckFullCertificateNotTrusted() throws Exception {
        // given
        String crlUrl = "https://localhost/crl";
        String revokedSerialFromList = "0011";
        CertificateFactory cf = CertificateFactory.getInstance("X.509");
        X509CRL crl = (X509CRL) cf.generateCRL(getClass().getResourceAsStream("/certificates/smp-crl-test.crl"));

        Mockito.doReturn(crl).when(crlVerifierService).getCRLByURL(crlUrl);

        Calendar from = Calendar.getInstance();
        Calendar to = Calendar.getInstance();
        to.add(Calendar.DAY_OF_YEAR, 1);
        from.add(Calendar.DAY_OF_YEAR, -2);
        X509Certificate certificate = X509CertificateTestUtils.createX509CertificateForTest(
                revokedSerialFromList, S_SUBJECT_PEPPOL_NOT_TRUSTED, S_SUBJECT_PEPPOL_NOT_TRUSTED,
                OffsetDateTime.now().minusDays(2),
                OffsetDateTime.now().plusYears(1),
                Collections.singletonList(crlUrl));

        // when
        CertificateNotTrustedException result = assertThrows(CertificateNotTrustedException.class, () ->
                testInstance.checkFullCertificateValidity(certificate));
        //then
        assertThat(result.getMessage(), CoreMatchers.containsString("Certificate is not trusted!"));
    }


    @Test
    void testCheckFullCertificateValidityRevoked() throws Exception {
        // given
        String crlUrl = "https://localhost/crl";
        String revokedSerialFromList = "0011";
        CertificateFactory cf = CertificateFactory.getInstance("X.509");
        X509CRL crl = (X509CRL) cf.generateCRL(getClass().getResourceAsStream("/certificates/smp-crl-test.crl"));

        Mockito.doReturn(crl).when(crlVerifierService).downloadCRL(ArgumentMatchers.eq(crlUrl), ArgumentMatchers.anyBoolean());

        String certSubject = "CN=SMP Test,OU=eDelivery,O=DIGITAL,C=BE";
        X509Certificate certificate = X509CertificateTestUtils.createX509CertificateForTest(
                revokedSerialFromList, certSubject, certSubject,
                OffsetDateTime.now().minusDays(2),
                OffsetDateTime.now().plusYears(1),
                Collections.singletonList(crlUrl));
        // add as trusted certificate
        testInstance.addCertificate(UUID.randomUUID().toString(), certificate);

        // when
        CertificateRevokedException result = assertThrows(CertificateRevokedException.class, () ->
                testInstance.checkFullCertificateValidity(certificate));
        //then
        assertThat(result.getMessage(), CoreMatchers.containsString("Certificate has been revoked"));
    }

    @Test
    void testCheckFullCertificateValidityInvalidKey() throws Exception {
        // given
        String certSubject = "CN=SMP Test,OU=eDelivery,O=DIGITAL,C=BE";
        X509Certificate certificate = X509CertificateTestUtils.createX509CertificateForTest(
                null, certSubject, certSubject,
                OffsetDateTime.now().minusDays(2),
                OffsetDateTime.now().plusYears(1),
                Collections.emptyList()
        );
        Mockito.doReturn(Collections.singletonList("InvalidKeyTest")).when(configurationService).getAllowedCertificateKeyTypes();
        // add as trusted certificate
        testInstance.addCertificate(UUID.randomUUID().toString(), certificate);

        CertificateException result = assertThrows(CertificateException.class, () ->
                testInstance.checkFullCertificateValidity(certificate));

        //then
        assertThat(result.getMessage(), CoreMatchers.containsString("Certificate does not have allowed key algorithm type!"));
    }

    @Test
    void testCheckFullCertificateValidityNotForceCRL() throws Exception {
        // given
        String crlUrl = "https://localhost/crl";
        String revokedSerialFromList = "0011";
        CertificateFactory cf = CertificateFactory.getInstance("X.509");
        X509CRL crl = (X509CRL) cf.generateCRL(getClass().getResourceAsStream("/certificates/smp-crl-test.crl"));
        Mockito.doReturn(false).when(configurationService).forceCRLValidation();
        Mockito.doThrow(new SMPRuntimeException(ErrorCode.CERTIFICATE_ERROR, "Error occurred while downloading CRL:" + crlUrl, "")).when(crlVerifierService).downloadURL(crlUrl);

        String certSubject = "CN=SMP Test,OU=eDelivery,O=DIGITAL,C=BE";
        X509Certificate certificate = X509CertificateTestUtils.createX509CertificateForTest(
                revokedSerialFromList, certSubject, certSubject,
                OffsetDateTime.now().minusDays(2),
                OffsetDateTime.now().plusYears(1),
                Collections.singletonList(crlUrl));
        // add as trusted certificate
        testInstance.addCertificate(UUID.randomUUID().toString(), certificate);


        //then should be thrown CertificateRevokedException but is not
        // when
        testInstance.checkFullCertificateValidity(certificate);
    }

    @Test
    void testCheckFullCertificateValidityOK() throws Exception {
        // given
        String crlUrl = "https://localhost/crl";
        CertificateFactory cf = CertificateFactory.getInstance("X.509");
        X509CRL crl = (X509CRL) cf.generateCRL(getClass().getResourceAsStream("/certificates/smp-crl-test.crl"));

        Mockito.doReturn(crl).when(crlVerifierService).downloadCRL(crlUrl, true);

        String certSubject = "CN=SMP Test,OU=eDelivery,O=DIGITAL,C=BE";
        X509Certificate certificate = X509CertificateTestUtils.createX509CertificateForTest(certSubject);
        // add as trusted certificate
        testInstance.addCertificate(UUID.randomUUID().toString(), certificate);

        // when
        testInstance.checkFullCertificateValidity(certificate);

        // then
        //no errors should be thrown
    }

    @Test
    void testCreateAliasForCert() throws Exception {
        // given
        String certSubject = "CN=SMP Test,OU=eDelivery,O=DIGITAL,C=BE";
        X509Certificate certificate = X509CertificateTestUtils.createX509CertificateForTest(certSubject);
        // when
        String alias = testInstance.createAliasFromCert(certificate, null);

        // then
        assertEquals("smp test", alias);
    }


    @Test
    void testCreateAliasFoMultiValuerCert() throws Exception {
        // given
        String certSubject = "GIVENNAME=John+SERIALNUMBER=1+CN=SMP Test,OU=eDelivery,O=DIGITAL,C=BE";
        X509Certificate certificate = X509CertificateTestUtils.createX509CertificateForTest(certSubject);
        // when
        String alias = testInstance.createAliasFromCert(certificate, null);

        // then
        assertEquals("smp test", alias);
    }

    @Test
    void testValidateCertificatePolicyLegacyMatchOk() throws Exception {
        String certSubject = "CN=SMP Test,OU=eDelivery,O=DIGITAL,C=BE";
        X509Certificate certificate = X509CertificateTestUtils.createX509CertificateForTest(certSubject, BigInteger.TEN, Collections.singletonList(CERTIFICATE_POLICY_QCP_NATURAL));
        Mockito.doReturn(Arrays.asList(CERTIFICATE_POLICY_QCP_LEGAL, CERTIFICATE_POLICY_QCP_NATURAL)).when(configurationService).getAllowedCertificatePolicies();
        testInstance.validateCertificatePolicyMatchLegacy(certificate);
    }

    @Test
    void testValidateCertificatePolicyLegacyMatchEmpty() throws Exception {
        String certSubject = "CN=SMP Test,OU=eDelivery,O=DIGITAL,C=BE";
        X509Certificate certificate = X509CertificateTestUtils.createX509CertificateForTest(certSubject, BigInteger.TEN, null);
        Mockito.doReturn(Arrays.asList(CERTIFICATE_POLICY_QCP_LEGAL, CERTIFICATE_POLICY_QCP_NATURAL)).when(configurationService).getAllowedCertificatePolicies();

        CertificateException result = assertThrows(CertificateException.class,
                () -> testInstance.validateCertificatePolicyMatchLegacy(certificate));
        assertThat(result.getMessage(), CoreMatchers.startsWith("Certificate has empty CertificatePolicy extension. Certificate:"));
    }

    @Test
    void testValidateCertificatePolicyLegacyMatchMismatch() throws Exception {
        String certSubject = "CN=SMP Test,OU=eDelivery,O=DIGITAL,C=BE";
        X509Certificate certificate = X509CertificateTestUtils.createX509CertificateForTest(certSubject, BigInteger.TEN, Collections.singletonList(CERTIFICATE_POLICY_QCP_LEGAL_QSCD));
        Mockito.doReturn(Arrays.asList(CERTIFICATE_POLICY_QCP_LEGAL, CERTIFICATE_POLICY_QCP_NATURAL)).when(configurationService).getAllowedCertificatePolicies();

        CertificateException result = assertThrows(CertificateException.class,
                () -> testInstance.validateCertificatePolicyMatchLegacy(certificate));
        assertThat(result.getMessage(), CoreMatchers.startsWith("Certificate policy verification failed."));
    }

    @Test
    void validateCertificateNotUsed() throws CertificateException {
        String certId = "cn=test" + UUID.randomUUID() + ",o=test,c=eu:123456";
        CertificateRO certificateRO = new CertificateRO();
        certificateRO.setCertificateId(certId);
        // when
        testInstance.validateCertificateNotUsed(certificateRO);
        //then no error is thrown because

    }

}
